import time
from datetime import datetime
from threading import Thread
from dateutil.relativedelta import relativedelta
from dateutil.rrule import *
from dateutil import parser

class Threader(Thread):
	def __init__(self, job):
		super().__init__()
		self.job = job

	def run(self):
		try:
			if self.job.seconds_to_run > 0:
				time.sleep(self.job.seconds_to_run)
			if self.job.args and self.job.kwargs:
				self.job.function(*self.job.args, **self.job.kwargs)
			elif self.job.args and not self.job.kwargs:
				self.job.function(*self.job.args)
			elif not self.job.args and self.job.kwargs:
				self.job.function(**self.job.kwargs)
			else:
				self.job.function()
			self.job.run_count += 1
		except:
			self.job.error_count += 1
			print('There was an error in', self.job.name)
			raise

class Job(object):
	function = None
	id = None
	name = None
	args = None
	kwargs = None
	thread = None
	job_type = None
	status = None
	seconds_to_run = None
	retrys = None
	next_run_time = None
	how_many = None
	done = False
	run_count = 0
	error_count = 0

	def __init__(self, func=None, args=None, kwargs=None, job_type=None,
	             status='new', retrys=3, how_many=float('inf')):
		self.function = func
		self.id = id(func)
		self.name = func.__name__
		self.args = args
		self.kwargs = kwargs
		self.job_type = job_type
		self.status = status
		self.retrys = retrys
		self.how_many = how_many

class RecuringJob(Job):
	interval = None
	start_datetime = None
	days_of_week = None

	def __init__(self, func=None, args=None, kwargs=None, job_type=None,
	             status='new', interval=DAILY, start_datetime=datetime.now(),
	             days_of_week=None, retrys=3, how_many=float('inf')):
		super().__init__(func=func, args=args, kwargs=kwargs, job_type=job_type,
		        status=status, retrys=retrys, how_many=how_many)
		self.interval = interval
		self.start_datetime = start_datetime
		self.days_of_week = days_of_week

	def find_next_time(self):
		start_datetime = self.start_datetime
		time_delta = start_datetime - datetime.now()
		if time_delta.days < 0:
			if self.interval == SECONDLY:
				start_datetime = datetime.now()
			elif self.interval == MINUTELY:
				start_datetime = start_datetime.replace(datetime.now().year, datetime.now().month,
						datetime.now().day, datetime.now().hour, datetime.now().minute)
			elif self.interval == HOURLY:
				start_datetime = start_datetime.replace(datetime.now().year, datetime.now().month,
						datetime.now().day, datetime.now().hour)
			elif self.interval == DAILY:
				start_datetime = start_datetime.replace(datetime.now().year, datetime.now().month,
						datetime.now().day)
			elif self.interval == MONTHLY:
				start_datetime = start_datetime.replace(datetime.now().year, datetime.now().month)
			elif self.interval == YEARLY:
				start_datetime = start_datetime.replace(datetime.now().year)
		self.start_datetime = start_datetime
		runs = list(rrule(self.interval, dtstart=self.start_datetime, count=2))
		self.seconds_to_run = (runs[0] - datetime.now()).total_seconds()
		self.next_run_time = runs[1]

	def schedule_next(self):
		if self.error_count < self.retrys:
			if self.run_count < self.how_many:
				if self.status == 'new':
					self.find_next_time()
					self.thread = Threader(self)
					self.thread.start()
					self.status = 'waiting'
				else:
					if not self.thread.is_alive():
						self.seconds_to_run = (self.next_run_time - datetime.now()).total_seconds()
						if self.seconds_to_run > 0:
							self.thread = Threader(self)
							self.thread.start()
						self.find_next_time()
			else:
				self.done = True
		else:
			self.done = True

class DelayedJob(Job):
	time_delta = None
	years = None
	months = None
	days = None
	hours = None
	minutes = None
	seconds = None

	def __init__(self, func=None, args=None, kwargs=None, job_type=None,
	             status='new',  time_delta=None, years=None, months=None,
	             days=None, hours=None, minutes=None, seconds=None,
	             retrys=3, how_many=float('inf')):
		super().__init__(func=func, args=args, kwargs=kwargs, job_type=job_type,
		        status=status, retrys=retrys, how_many=how_many)
		self.time_delta = time_delta
		self.years = years
		self.months = months
		self.days = days
		self.hours = hours
		self.minutes = minutes
		self.seconds = seconds

	def schedule_next(self):
		if self.error_count < self.retrys:
			if self.run_count < self.how_many:
				if self.status == 'new':
					self.time_delta = datetime.now() + relativedelta(years=self.years,
							months=self.months, days=self.days, hours=self.hours,
							minutes=self.minutes, seconds=self.seconds) - datetime.now() #this may look strange but it creates a timedelta object (datetime + relativedelta = new_datetime; new_datetime - datetime = timedelta)
					self.seconds_to_run = self.time_delta.total_seconds()
					self.thread = Threader(self)
					self.thread.start()
					self.status = 'waiting'
				else:
					if not self.thread.is_alive():
						self.time_delta = datetime.now() + relativedelta(years=self.years,
								months=self.months, days=self.days, hours=self.hours,
								minutes=self.minutes, seconds=self.seconds) - datetime.now()
						self.seconds_to_run = self.time_delta.total_seconds()
						self.thread = Threader(self)
						self.thread.start()
			else:
				self.done = True
		else:
			self.done = True

class Scheduler:
	job_store = []

	def __init__(self, log=None):
		self.log = log or object
		def event_loop():
			while True:
				start = time.clock()
				for job in self.job_store:
					if job.done:
						self.job_store.remove(job)
					else:
						job.schedule_next()
				time.sleep(.5 - (time.clock() - start))  #if debuging take out the time.clock() - start
		Thread(target=event_loop).start()

	def schedule_recuring_job(self, interval='daily',
	        start_date=datetime.now().strftime('%Y-%m-%d'),
	        start_time='00:00:00', days_of_week=None,
	        how_many=float('inf'), retrys=3):
		def wrap(f):
			def wrapped(*args, **kwargs):
				Scheduler.job_store.append(RecuringJob(f, args=args, kwargs=kwargs,
				        job_type='recuring', interval=self.get_interval(interval),
				        start_datetime=parser.parse('%s %s' % (start_date, start_time),
				        yearfirst=True, ignoretz=True), days_of_week=days_of_week,
				        how_many=how_many, retrys=retrys))
			return wrapped()
		return wrap

	def add_recuring_job(self, func, args=None, kwargs=None,
	         interval='daily', start_date=datetime.now().strftime('%Y-%m-%d'),
	        start_time='00:00:00', days_of_week=None, how_many=float('inf'),
	        retrys=3):
		self.job_store.append(RecuringJob(func=func, args=args, kwargs=kwargs,
				job_type='recuring', interval=self.get_interval(interval),
				start_datetime=parser.parse('%s %s' % (start_date, start_time),
				yearfirst=True, ignoretz=True), days_of_week=days_of_week,
				how_many=how_many, retrys=retrys))

	def get_interval(self, interval):
		interval = interval.lower()
		if interval == 'secondly':
			interval = SECONDLY
		elif interval == 'minutely':
			interval = MINUTELY
		elif interval == 'hourly':
			interval = HOURLY
		elif interval == 'daily':
			interval = DAILY
		elif interval == 'weekly':
			interval = WEEKLY
		elif interval == 'monthly':
			interval = MONTHLY
		elif interval == 'yearly':
			interval = YEARLY
		else:
			raise TypeError
		return interval

	def schedule_delayed_job(self, years=0, months=0, days=0, hours=0,
	        minutes=0, seconds=0, how_many=float('inf'), retrys=3):
		def wrap(f):
			def wrapped(*args, **kwargs):
				Scheduler.job_store.append(DelayedJob(func=f, args=args, kwargs=kwargs,
						job_type='delayed', years=years, months=months, days=days,
						hours=hours, minutes=minutes, seconds=seconds,
						how_many=how_many, retrys=retrys))
			return wrapped()
		return wrap

	def add_delayed_job(self, func, args=None, kwargs=None, years=0,
	        months=0, days=0, hours=0, minutes=0, seconds=0,
	        how_many=float('inf'), retrys=3):
		self.job_store.append(DelayedJob(func=func, args=args, kwargs=kwargs,
		        job_type='delayed', years=years, months=months, days=days,
		        hours=hours, minutes=minutes, seconds=seconds,
		        how_many=how_many, retrys=retrys))